"""
This script is for 
- split test set into proxy and test set
- partition train set into multiple clients
- get optimal budget for proxy set
- train budget prediction model
- use budget prediction model in testing time, to predict the budget allocation for each client per query
"""

import torch

from datasets import concatenate_datasets, load_dataset, DatasetDict, Dataset
from accelerate import Accelerator
from rank_bm25 import BM25Okapi
import numpy as np
from tqdm import trange
from nltk.tokenize import word_tokenize

from openicl.icl_retriever import BaseRetriever, BM25Retriever
from openicl.utils.check_type import _check_str
from openicl import DatasetReader, PromptTemplate, PPLInferencer, AccEvaluator

import argparse
import json
import os
import wandb
import socket
import copy

from util.template import gen_template_classification
from util.retriever import get_retriever, get_fedretriever, get_server_retriever
from util.partition import (
    cls_iid_partition,
    cls_noniid_partition,
    datasplit_subset,
    data2usename,
    data2numcls,
)

from util.icl_ppl_fed_budgetmodel_inferencer import PPLFedBudgetModelInferencer
from util.misc import setup_seed


def print_score(
    args,
    output_json_path,
):
    pred_file = os.path.join(
        output_json_path,
        f"prediction_{'debug' if args.debug else 'run'}.json",
    )

    label_file = os.path.join(
        output_json_path,
        f"label_{'debug' if args.debug else 'run'}.json",
    )
    # load
    with open(pred_file, "r") as f:
        prediction = data_loaded = json.load(f)
    with open(label_file, "r") as f:
        reference = data_loaded = json.load(f)

    evaluator = AccEvaluator()
    score = evaluator.score(predictions=prediction, references=reference)
    print(score)
    return score


def run(args):
    # ===== 1. Data Setting Preparation =====
    # load dataset
    dataset = load_dataset(data2usename[args.dataset])

    if "test" not in dataset:
        test_split = "validation"
    else:
        test_split = "test"

    # slice a subset for proxy dataset on server
    if args.proxy_split is not None:
        proxy_data, remain_data = datasplit_subset(
            dataset[args.proxy_split],
            subset_num=args.proxy_size,
            split=args.proxy_split,
            verbose=True,
            return_remain=True,
        )
        dataset[args.proxy_split] = remain_data

    # partition data into clients
    if args.partition == "iid":
        fed_dataset = cls_iid_partition(
            dataset=dataset,
            split="train",
            data_name=args.dataset,
            num_clients=args.num_clients,
            test_split=test_split,
            subset_num=args.subset_num,
        )
    elif args.partition == "noniid":
        fed_dataset = cls_noniid_partition(
            dataset=dataset,
            split="train",
            data_name=args.dataset,
            major_classes_num=args.major_classes_num,
            num_clients=args.num_clients,
            test_split=test_split,
            subset_num=args.subset_num,
        )
    else:
        raise ValueError(
            f"args.partition can only be 'iid' or 'noniid', rather than '{args.partition}'."
        )

    # save selected test query subset
    subset_orig_idxs = fed_dataset[test_split]["idx"]
    subset_orig_idxs_file = os.path.join(args.log_dir, "query_subset_orig_idxs.json")
    with open(subset_orig_idxs_file, "w") as f:
        json.dump(subset_orig_idxs, f)

    # add client idx for each local training sample (used in proxy set's optimal budget allocation process)
    for cid in range(args.num_clients):
        local_sample_num = len(fed_dataset[f"train-client{cid}"])
        cid_col = [cid for _ in range(local_sample_num)]
        fed_dataset[f"train-client{cid}"] = fed_dataset[
            f"train-client{cid}"
        ].add_column("cid", copy.deepcopy(cid_col))
    print("Add client ID column for each local training dataset.")

    # make datareader for all clients' local training data
    fed_reader = []
    for cid in range(args.num_clients):
        fed_reader.append(
            DatasetReader(
                fed_dataset[f"train-client{cid}"],
                input_columns=["sentence"],
                output_column="label",
            )
        )

    # get template for classification
    tp_dict = gen_template_classification(args)
    template = PromptTemplate(tp_dict, {"sentence": "</text>"}, ice_token="</E>")

    # generate retriever for each client: each local train data will be used to build local train corpus
    fed_retrievers = []
    RETRIEVER = get_fedretriever(args)
    for cid in range(args.num_clients):
        retriever = RETRIEVER(fed_reader[cid], ice_num=args.local_ice_num)
        fed_retrievers.append(retriever)

    #  ===== 2. Obtain the optimal budget allocation for proxy dataset =====
    setup_seed(args.seed)
    inferencer = PPLFedBudgetModelInferencer(
        model_name=args.model, output_json_filepath=args.log_dir, args=args
    )

    # ===== 3. Obtain the optimal budget allocation for proxy dataset =====
    ice_client_source, opt_per_client_budget, res_list = (
        inferencer.opt_budget_allocation(
            fed_retrievers,
            query_dataset=fed_dataset[test_split],
            ice_template=template,
            output_json_filepath=args.log_dir,
            local_ice_num=args.local_ice_num,
            server_ice_num=args.server_ice_num,
            prefix=test_split,
            args=args,
        )
    )

    if args.inference:
        # re-format the optimal budgets for each client
        per_query_budgets_list = []
        for res in res_list:
            q_orig_idx = res["idx"]
            per_query_budgets_list.append(opt_per_client_budget[q_orig_idx])
        opt_query_budgets = [[*content] for content in zip(*per_query_budgets_list)]

        # ------ inference using optimal budgets
        predictions = inferencer.inference(
            retrievers=fed_retrievers,
            query_dataset=fed_dataset[test_split],
            ice_template=template,
            output_json_filepath=args.log_dir,
            concat=args.concat,
            args=args,
            use_budget_model=False,
            query_budgets=opt_query_budgets,
        )

        # ----- save inference prediction
        prediction_file = os.path.join(
            args.log_dir,
            f"prediction_{'debug' if args.debug else 'run'}.json",
        )
        with open(prediction_file, "w") as f:
            json.dump(predictions, f)

        label_file = os.path.join(
            args.log_dir,
            f"label_{'debug' if args.debug else 'run'}.json",
        )
        with open(label_file, "w") as f:
            # json.dump(data.references, f)
            json.dump(fed_dataset["test"]["label"], f)


if __name__ == "__main__":
    args = argparse.ArgumentParser()
    args.add_argument("--dataset", type=str, default="sst2", help="Dataset name")
    args.add_argument(
        "--subset_num",
        type=int,
        default=None,
        help="Number of subset test set for query",
    )
    args.add_argument("--proxy_split", default=None, type=str)
    args.add_argument("--proxy_size", default=500, type=int)
    args.add_argument("--partition", default="iid", choices=["iid", "noniid"], type=str)
    args.add_argument("--num_clients", type=int, default=3)
    args.add_argument("--major_classes_num", default=-1, type=int)
    args.add_argument(
        "--model",
        type=str,
        default="EleutherAI/gpt-neo-2.7B",
        help="Pretrained LLM model name",
    )
    args.add_argument("--local_ice_num", default=None, type=int)
    args.add_argument(
        "--server_ice_num",
        default=-1,
        type=int,
        help="Server side ICE Number, server_ice_num <= num_clients * local_ice_num",
    )
    args.add_argument(
        "--concat", default="reorder", type=str, choices=["simple", "merge", "reorder"]
    )
    args.add_argument(
        "--retriever", type=str, default="topk", help="Server side Retriever Type"
    )  # use 'bm25'
    args.add_argument("--inference", action="store_true")

    args.add_argument(
        "--log_dir",
        type=str,
        default=f"fed_icl_log",
        help="Logging directory",
    )
    args.add_argument("--debug", action="store_true")
    args.add_argument("--run", action="store_true")
    args.add_argument("--seed", default=0, type=int)
    args.add_argument("--proj_name", default="FL-ICL-debug")
    args.add_argument("--group_name", default=None, type=str)

    args = args.parse_args()

    args.num_classes = data2numcls[args.dataset]

    # prepare output files
    model = args.model.replace("/", "_")
    part_name = f"{args.partition}_clients={args.num_clients}"
    if args.partition == "noniid":
        part_name += f"_majorclass={args.major_classes_num}"

    data_folder = f"{args.dataset}"
    if args.subset_num is not None:
        data_folder += f"_query-num={args.subset_num}"

    if args.proxy_split is not None:
        data_folder += f"_proxy={args.proxy_split}-{args.proxy_size}"

    host_name = socket.gethostname()
    if "server_name" in host_name.lower():
        cache_root = "cache/root/directory"
    else:
        raise ValueError(f"Check the server hostname for log_dir initialization.")

    # add budget model info in log_dir
    args.log_dir = os.path.join(
        cache_root,
        args.log_dir,
        f"{data_folder}/{part_name}/oracle_budget_model={model}_retriever=fed{args.retriever}_local-ice={args.local_ice_num}_server-ice={args.server_ice_num}/seed={args.seed}",
    )
    os.makedirs(args.log_dir, exist_ok=True)

    # perform FL-ICL pipeline
    if args.run:
        run_name = f"oraclebudget_fed_{data_folder}_{part_name}_model={model}_retriever=fed{args.retriever}_local-ice={args.local_ice_num}_server-ice={args.server_ice_num}_seed={args.seed}"
        wb_run = wandb.init(
            config=args, project=args.proj_name, name=run_name, group=args.group_name
        )

    if args.run:
        run(args)
        if args.inference:
            scores = print_score(args, args.log_dir)
            wb_run.log(scores)

    wb_run.finish()
